# -*- coding: utf-8 -*-
#
# Copyright 2018 Data61, CSIRO
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools

DESCRIPTION = "Python library for machine learning on graphs"
URL = "https://github.com/stellargraph/stellargraph"
VERSION = "0.4.1"

# Required packages
REQUIRES = [
    "keras>=2.2.0",
    "tensorflow>=1.8",
    "numpy>=1.14, <1.15",
    "networkx>=2.1",
    "scikit_learn>=0.18",
    "matplotlib>=2.2",
    "gensim>=3.4.0",
    "pandas>=0.23"
]

EXTRAS_REQURES = {"demos": ["numba"], "test": ["pytest", "pandas", "pytest-benchmark"]}

# Long description
with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="stellargraph",
    version=VERSION,
    description=DESCRIPTION,
    author="Data61, CSIRO",
    author_email="stellar.admin@csiro.au",
    url=URL,
    license="Apache 2.0",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    include_package_data=True,
    python_requires='>=3.5.0, <3.7.0',
    install_requires=REQUIRES,
    extras_require=EXTRAS_REQURES,
    packages=setuptools.find_packages(exclude=("tests",)),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2018 Data61, CSIRO
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["train_val_test_split", "NodeSplitter"]

import numpy as np
import pandas as pd
from stellargraph.core.graph import StellarGraphBase
from stellargraph import globalvar


# Easier functional interface for the splitter:
def train_val_test_split(
    G,
    node_type=None,
    test_size=0.4,
    train_size=0.2,
    targets=None,
    split_equally=False,
    seed=None,
):
    """
        Splits node data into train, test, validation, and unlabeled sets.

        Any nodes that have a target value equal to globals.UNKNOWN_TARGET_ATTRIBUTE are added to the unlabeled set.

        The validation set includes all nodes that remain after the train, test and unlabeled sets have been
        created. As a result, it is possible the the validation set is empty.

    Args:
        G : StellarGraph containing the nodes to be split.

        test_size: float, int
            If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to
            include in the test split. If int, represents the absolute number of test samples.

        train_size: float, int
            If float, should be between 0.0 and 1.0 and represent the proportion of the dataset to
            include in the train split. If int, represents the absolute number of train samples.
            If None, the value is automatically set to the complement of the test size.

        seed: int or None, optional (default=None)
            If this is an int the seed will be used to initialize a random number generator,
            otherwith the numpy default will be used.

        shuffle : boolean, optional (default=True)
            Whether or not to shuffle the data before splitting. If shuffle=False then stratify must be None.

        targets: None or DataFrame or dict, optional (default=None)
            If False the nodes are randomly assigned to each partition. If this is a Pandas DataFrame
            with node_ids as the index, or a dictionary with node_ids as keys and target as values,
            then these values will be used to find unlabelled nodes and if `split_equally` is True
            these target values will be used to sample equal numbers of each class.

        split_equally: bool (default=False)
            if `split_equally` is True the values passed into the targets argument will be used to
            sample equal numbers for the train split by class label.

    Returns:
            y_train, y_val, y_test, y_unlabeled
    """
    node_splitter = NodeSplitter()

    # Get list of nodes to split
    if node_type is None:
        nodes = list(G)

    elif isinstance(G, StellarGraphBase):
        nodes = G.nodes_of_type(node_type)

    else:
        raise TypeError("G must be a StellarGraph is node_type is not None")

    # Number of nodes and number without a label
    n_nodes = len(nodes)

    # Extract the target information
    if targets is not None:
        # TODO: The equal sampling option will fail if these values are not hashable.
        # Check that split_equally_by_target is the correct type
        if isinstance(targets, pd.DataFrame):
            target_values = [
                targets.loc[n]
                if n in targets.index
                else globalvar.UNKNOWN_TARGET_ATTRIBUTE
                for n in nodes
            ]

        elif isinstance(targets, dict):
            target_values = [
                targets.get(n, globalvar.UNKNOWN_TARGET_ATTRIBUTE) for n in nodes
            ]

        else:
            raise TypeError(
                "The targets are expected to be either a Pandas DataFrame or a dict."
            )

        n_known = sum(t != globalvar.UNKNOWN_TARGET_ATTRIBUTE for t in targets)

    else:
        n_known = n_nodes
        target_values = [0] * n_nodes

    if n_known == 0:
        raise RuntimeError("No nodes with target attribute to split.")

    # Find the number of nodes to use in the training set
    if isinstance(train_size, float) and (0 < train_size <= 1):
        train_size_n = int(n_known * train_size)

    elif isinstance(train_size, int):
        train_size_n = train_size

    else:
        raise ValueError("Splitter: train_size should be specified as a float or int")

    # Find the number of nodes to use in the test set
    if isinstance(test_size, float) and (0 < test_size <= 1):
        test_size_n = int(n_known * test_size)

    elif isinstance(test_size, int):
        test_size_n = test_size

    else:
        raise ValueError("Splitter: train_size should be specified as a float or int")

    # Find the number of nodes to use in the validation set
    val_size = None
    if isinstance(val_size, float) and (0 < val_size <= 1):
        val_size_n = int(n_known * val_size)

    elif isinstance(val_size, int):
        val_size_n = val_size

    else:
        val_size_n = max(0, n_known - (train_size_n + test_size_n))

    # Check that these sizes make sense
    if (train_size_n + test_size_n + val_size_n) > n_known:
        raise ValueError(
            "Number of train, test and val nodes "
            "is greater than the total number of labelled nodes."
        )

    # Now the splitter needs the node IDs and labels zipped together
    # TODO: This is a hack as the splitter only works when this array is sting type
    nodeid_and_label = np.array([nl for nl in enumerate(target_values)], dtype="U")

    # If stratified sampling, we need the target labels.
    if split_equally and targets is not None:
        class_set = set(target_values)

        # Remove the unknown target type
        class_set.discard(globalvar.UNKNOWN_TARGET_ATTRIBUTE)

        # The number of classes we have
        n_classes = len(class_set)

        if n_classes == 0:
            raise RuntimeError(
                "Found no usable target classes in split_equally_by_targets."
            )

        if train_size_n < n_classes:
            raise RuntimeError(
                "The number of classes must be smaller than the training size."
            )

        # The number of nodes we want per class
        p = int(train_size_n / n_classes)

        splits = node_splitter.train_test_split(
            y=nodeid_and_label, p=p, method="count", test_size=test_size_n, seed=seed
        )

    else:
        splits = node_splitter.train_test_split(
            y=nodeid_and_label,
            method="absolute",
            train_size=train_size_n,
            test_size=test_size_n,
            seed=seed,
        )

    # Get the node_ids out of the splitter
    node_ids_out = [[nodes[int(ind)] for ind in split[:, 0]] for split in splits]

    return node_ids_out


class NodeSplitter(object):
    def __init__(self):
        self.format_epgm = False
        self.g_epgm = None
        self.g_id = None
        self._random = None

    def _get_nodes(self, graph_nodes, node_type, target_attribute):
        """
        Returns a list of node IDs for the subset of graph_nodes that have the given node type.

        Args:
            graph_nodes: <list> List of OrderedDict with vertex data for graph in EPGM format
            node_type: <str> The node type of interest
            target_attribute: <str> The target attribute key

        Returns:
            <list> List of node IDs that have given node type

        """
        # This code will fail if a node of node_type is missing the target_attribute.
        # We can fix this by using node['data'].get(target_attribute, None) so that at least all nodes of the
        # given type are returned. However, we must check for None in target_attribute later to exclude these nodes
        # from being added to train, test, and validation datasets.
        y = [
            (
                node["id"],
                node["data"].get(target_attribute, globalvar.UNKNOWN_TARGET_ATTRIBUTE),
            )
            for node in graph_nodes
            if node["meta"][globalvar.TYPE_ATTR_NAME] == node_type
        ]

        return y

    def _check_parameters(self, y, p, method, test_size, train_size, seed):
        """
        Checks that the parameters have valid values. It not, then it raises a ValueError exception with a
        message corresponding to the invalid parameter.

        Args:
            y: <numpy array> Array of size Nx2 containing node id, label columns.
            p: <int or float> Percent or count of the number of points for each class to sample.
            method: <str> One of 'count', 'percent', or 'absolute'.
            test_size: <int> number of points in the test set. For method 'count', it should be less than or equal to
            N - (np.unique(labels) * nc) where N is the number of labeled points in y.
            train_size: <int> The number of points in the train set only used by method 'absolute'.
            seed: <int> seed for random number generator, positive int or 0

        """
        if not isinstance(y, np.ndarray):
            raise ValueError("({}) y should be numpy array".format(type(self).__name__))

        if method != "count" and method != "percent" and method != "absolute":
            raise ValueError(
                "({}) Valid methods are 'count', 'percent', and 'absolute' not {}".format(
                    type(self).__name__, method
                )
            )

        if seed is not None:
            if seed < 0:
                raise ValueError(
                    "({}) The random number generator seed value, seed, should be positive integer or None.".format(
                        type(self).__name__
                    )
                )
            if not isinstance(seed, int):
                raise ValueError(
                    "({}) The random number generator seed value, seed, should be integer type or None.".format(
                        type(self).__name__
                    )
                )

        if method == "count":
            if not isinstance(p, int) or p <= 0:
                raise ValueError(
                    "({}) p should be positive integer".format(type(self).__name__)
                )
            if test_size is None or not isinstance(test_size, int) or test_size <= 0:
                raise ValueError(
                    "({}) test_size must be positive integer".format(
                        type(self).__name__
                    )
                )

        elif method == "percent":
            if not isinstance(p, float) or p < 0. or p > 1.:
                raise ValueError(
                    "({}) p should be float in the range [0,1].".format(
                        type(self).__name__
                    )
                )

        elif method == "absolute":
            if test_size is None or not isinstance(test_size, int) or test_size <= 0:
                raise ValueError(
                    "({}) test_size should be positive integer".format(
                        type(self).__name__
                    )
                )
            if train_size is None or not isinstance(train_size, int) or train_size <= 0:
                raise ValueError(
                    "({}) train_size should be positive integer".format(
                        type(self).__name__
                    )
                )

    def train_test_split(
        self, y=None, p=10, method="count", test_size=None, train_size=None, seed=None
    ):
        """
        Splits node data into train, test, validation, and unlabeled sets.

        Any points in y that have value globals.UNKNOWN_TARGET_ATTRIBUTE are added to the unlabeled set.

        The validation set includes all the point that remain after the train, test and unlabeled sets have been
        created. As a result, it is possible the the validation set is empty, e.g., when method is set to 'percent'.

        The train, and test sets are build based on the specified method, 'count', 'percent', or 'absolute'.

        method='count': The value of parameter p specifies the number of points in the train set for each class. The
        test set size must be specified using the test_size parameter.

        method='percent': The value of parameter p specifies the train set size (and 1-p the test set size) as a
        percentage of the total number of points in y (including the unlabeled points.) The split is performed uniformly
        at random and the point labels (as specified in y) are not taken into account.

        method='absolute': The values of the parameters train_size and test_size specify the size of the train and test
        sets respectively. Points are selected uniformly at random and the label (as specified in y) are not taken into
        account.

        Args:
            y: <numpy array> Array of size Nx2 containing node id, label columns.
            p: <int or float> Percent or count of the number of points for each class to sample.
            method: <str> One of 'count', 'percent', or 'absolute'.
            test_size: <int> number of points in the test set. For method 'count', it should be less than or equal to
            N - (np.unique(labels) * nc) where N is the number of labeled points in y.
            train_size: <int> The number of points in the train set only used by method 'absolute'.
            seed: <int> seed for random number generator, positive int or 0

        Returns:
            y_train, y_val, y_test, y_unlabeled
        """
        self._check_parameters(
            y=y,
            p=p,
            method=method,
            test_size=test_size,
            train_size=train_size,
            seed=seed,
        )

        if self._random is None:
            self._random = np.random.RandomState(seed=seed)

        if method == "count":
            return self._split_data(y, p, test_size)
        elif method == "percent":
            n_unlabelled_points = np.sum(y[:, 1] == globalvar.UNKNOWN_TARGET_ATTRIBUTE)
            train_size = int((y.shape[0] - n_unlabelled_points) * p)
            test_size = y.shape[0] - n_unlabelled_points - train_size
            return self._split_data_absolute(
                y=y, test_size=test_size, train_size=train_size
            )
        elif method == "absolute":
            return self._split_data_absolute(
                y=y, test_size=test_size, train_size=train_size
            )

    def _split_data_absolute(self, y, test_size, train_size):
        """
        Splits given data such that the sizes of the test and train sets are fixed to the values given.

        Args:
            y: <numpy.ndarray> Array of size N x 2 containing node id + labels.
            test_size: <int> number of points in test set.
            train_size: <int> The number of points in the train set.

        Returns:
            y_train, y_val, y_test, y_unlabeled

        """
        # The label column in y could include None type, that is point with no ground truth label. These, if any,will
        # be returned separately in y_unlabeled dataset
        y_used = np.zeros(y.shape[0])  # initialize all the points are available

        # indexes of points with no class label:
        ind = np.nonzero(y[:, 1] == globalvar.UNKNOWN_TARGET_ATTRIBUTE)
        y_unlabeled = y[ind]
        y_used[ind] = 1

        ind = np.nonzero(y_used == 0)  # unused points
        ind_sampled = self._random.choice(ind[0], train_size, replace=False)
        y_train = y[ind_sampled]
        # mark these as used to make sure that they are not sampled for the test set
        y_used[ind_sampled] = 1

        # now sample test_size points for the test set
        ind = np.nonzero(y_used == 0)  # indexes of points that are not in training set
        if len(ind[0]) < test_size:
            raise Exception(
                "Not enough nodes available for the test set: available {} nodes, needed {}. Aborting".format(
                    len(ind[0]), test_size
                )
            )
        ind_sampled = self._random.choice(ind[0], test_size, replace=False)
        y_test = y[ind_sampled]
        y_used[ind_sampled] = 1
        # print("y_test shape: ", y_test.shape)

        # Validation set
        # the remaining labeled points (if any) go into the validation set
        ind = np.nonzero(y_used == 0)
        y_val = y[ind[0]]
        # print("y_val shape:", y_val.shape)

        return y_train, y_val, y_test, y_unlabeled

    def _split_data(self, y, nc, test_size):
        """
        Splits the data according to the scheme in Yang et al, ICML 2016, Revisiting semi-supervised learning with graph
        embeddings.

        Args:
            y: <numpy.ndarray> Array of size N x 2 containing node id + labels.
            nc: <int> number of points from each class in train set.
            test_size: <int> number of points in test set; it should be less than or equal to
            N - (np.unique(labels) * nc).

        Returns:
            y_train, y_val, y_test, y_unlabeled
        """
        y_used = np.zeros(y.shape[0])  # initialize all the points are available

        # indexes of points with no class label:
        ind = np.nonzero(y[:, 1] == globalvar.UNKNOWN_TARGET_ATTRIBUTE)
        y_unlabeled = y[ind]
        y_used[ind] = 1

        y_train = None

        class_labels = np.unique(y[:, 1])
        ind = class_labels == globalvar.UNKNOWN_TARGET_ATTRIBUTE
        class_labels = class_labels[np.logical_not(ind)]

        if test_size > y.shape[0] - class_labels.size * nc:
            # re-adjust so that none of the training samples end up in the test set
            test_size = y.shape[0] - class_labels.size * nc

        for clabel in class_labels:
            # indexes of points with class label clabel:
            ind = np.nonzero(y[:, 1] == clabel)
            # select nc of these at random for the training set
            if ind[0].size <= nc:
                # too few labeled examples for class so use half for training and half for testing
                ind_selected = self._random.choice(
                    ind[0], ind[0].size // 2, replace=False
                )
            else:
                ind_selected = self._random.choice(ind[0], nc, replace=False)
            # mark these as used to make sure that they are not sampled for the test set:
            y_used[ind_selected] = 1
            if y_train is None:
                y_train = y[ind_selected]
            else:
                # print("y_train shape:", y_train.shape)
                # print("y[ind_selected] shape:", y[ind_selected].shape)
                y_train = np.vstack((y_train, y[ind_selected]))

        # now sample test_size points for the test set
        ind = np.nonzero(y_used == 0)  # indexes of points that are not in training set
        if len(ind[0]) < test_size:
            raise Exception(
                "Not enough nodes available for the test set: available {} nodes, needed {}. Aborting".format(
                    len(ind[0]), test_size
                )
            )
        ind_selected = self._random.choice(ind[0], test_size, replace=False)
        y_test = y[ind_selected]
        y_used[ind_selected] = 1
        # print("y_test shape: ", y_test.shape)

        # the remaining points (if any) go into the validation set
        ind = np.nonzero(y_used == 0)
        y_val = y[ind[0]]
        # print("y_val shape:", y_val.shape)

        return y_train, y_val, y_test, y_unlabeled



# -*- coding: utf-8 -*-
#
# Copyright 2017-2018 Data61, CSIRO
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import networkx as nx
from networkx.readwrite import json_graph
import os
import json
import uuid
import numpy as np
import chardet
import scipy.sparse as sp

import multiprocessing
from multiprocessing import Pool
from functools import partial

# from progressbar import ProgressBar, SimpleProgress
from time import sleep


def node_neighbours(v, edges):
    """Returns a list of neighbours of vertex v"""
    return (v, [e[1] for e in edges if e[0] == v])


def node_neighbours_extended(v, nodes, edges):
    """Returns a list of neighbours of vertex v"""
    nodes = np.array(nodes)
    edges = np.array(edges)
    mask = np.in1d(edges[:, 0], v)
    return (v, np.where(np.in1d(nodes, edges[mask, 1])), edges[mask, 1])


class EPGM(object):
    """EPGM class with converter methods to edgelist, adjacency matrix, etc."""

    def _progress(self, item_name, n, step, arg, i):
        """Display progress of a loop, e.g., list or dict comprehension statement"""
        if (i + 1) % step == 0:
            print(
                "{}> {} {} out of {} processed ({}%)".format(
                    "-" * int((i / n) * 80),
                    item_name,
                    i + 1,
                    n,
                    round((i + 1) / n * 100, 1),
                )
            )
        return arg

    def _progressbarUpdate(self, pbar, step, arg, i):
        """update progress bar"""
        if (i + 1) % step == 0:
            pbar.update(i + 1)
        return arg

    @classmethod
    def _nx_to_json(self, G):
        # Convert G to json:
        G_json = json_graph.node_link_data(G)
        G_json["graph"].update({"id": G.id})

        # Convert G_json['nodes'] to strings:
        for n in G_json["nodes"]:
            n["id"] = str(n["id"])

        # Fix source and target of G_json['links'] to actual node ids:
        for l in G_json["links"]:
            l["source"] = str(G_json["nodes"][l["source"]]["id"])
            l["target"] = str(G_json["nodes"][l["target"]]["id"])

        return G_json

    @classmethod
    def _json_to_epgm(self, G_json, node_attributes, node_labels):
        """Create G_epmg from G_json"""

        G_epgm = {
            "graphs": [OrderedDict()],
            "vertices": [OrderedDict(n) for n in G_json["nodes"]],
            "edges": [OrderedDict(e) for e in G_json["links"]],
        }

        # update G_epgm['graphs']:
        G_epgm["graphs"][0].update(
            OrderedDict(
                [
                    ("id", G_json["graph"]["id"]),
                    ("data", {}),
                    ("meta", {"label": G_json["graph"].get("name", "unnamed")}),
                ]
            )
        )

        # update G_epgm['vertices']:
        if node_labels is None:
            node_labels = [""] * len(G_epgm["vertices"])
        for ind, v in enumerate(G_epgm["vertices"]):
            if node_attributes is not None:
                data = {
                    str(k): str(v)
                    for k, v in node_attributes.iloc[ind].items()
                    if v != 0
                }  # extract non-zero node attributes (sparse node attributes)
            else:
                data = {}
            v.update(
                OrderedDict(
                    [
                        ("data", data),
                        (
                            "meta",
                            {
                                "label": str(node_labels[ind]),
                                "graphs": [G_json["graph"]["id"]],
                            },
                        ),
                    ]
                )
            )

        # update G_epgm['edges']:
        edges_key_order = ["id", "source", "target", "data", "meta"]
        for e in G_epgm["edges"]:
            e.update(
                OrderedDict(
                    [
                        ("id", uuid.uuid4().hex),
                        ("data", {}),
                        ("meta", {"label": "", "graphs": [G_json["graph"]["id"]]}),
                    ]
                )
            )
        # reorder keys in e to edges_key_order:
        G_epgm["edges"] = [
            OrderedDict((k, e[k]) for k in edges_key_order) for e in G_epgm["edges"]
        ]

        return G_epgm

    @classmethod
    def _reorder_keys(self, G):
        """Apply correct order of keys in self.G"""

        # Graphs:
        graphs_key_order = ["id", "data", "meta"]
        # reorder keys in g to graphs_key_order:
        G["graphs"] = [
            OrderedDict((k, g[k]) for k in graphs_key_order) for g in G["graphs"]
        ]

        # Vertices:
        vertices_key_order = ["id", "data", "meta"]
        # reorder keys in v to vertices_key_order:
        G["vertices"] = [
            OrderedDict((k, v[k]) for k in vertices_key_order) for v in G["vertices"]
        ]

        # Edges:
        edges_key_order = ["id", "source", "target", "data", "meta"]
        # reorder keys in e to edges_key_order:
        G["edges"] = [
            OrderedDict((k, e[k]) for k in edges_key_order) for e in G["edges"]
        ]

        return G

    @classmethod
    def load(self, path):
        """Load graphs from EPGM path"""
        if not os.path.isdir(path):
            raise Exception("Path {} does not exist!".format(path))

        G_epgm = {"graphs": list(), "vertices": list(), "edges": list()}

        for k in G_epgm.keys():
            # detect the codec:
            fname = os.path.join(path, str(k) + ".json")
            # with open(fname, 'rb') as fb:  # open the file for reading in binary format
            #     line = fb.readline()   # read the 1st line and evalulate its length
            #     fb.seek(0)   # return to the start of the file
            #     enc = chardet.detect(fb.read(1000*len(line)))['encoding']
            # detect the encoding from up to 1000 first lines in the file

            enc = "utf-8"  # just use 'utf-8' for all files
            with open(fname, "r", encoding=enc) as fp:
                print("...reading {} using {} encoding...".format(fp.name, enc))
                lines = fp.readlines()

            if isinstance(G_epgm[k], list):
                for l in lines:
                    G_epgm[k].append(json.loads(l))
            else:
                raise Exception(
                    "type(G[", k, "]): unexpected type", type(G_epgm[k]), ", stopping."
                )

        G_epgm = self._reorder_keys(G=G_epgm)
        return G_epgm

    def node_types(self, graph_id):
        """List node labels(types) in graph with graph_id"""
        if not any([graph_id in g["id"] for g in self.G["graphs"]]):
            raise Exception("Graph with id {} does not exist".format(graph_id))

        node_types = [
            v["meta"]["label"]
            for v in self.G["vertices"]
            if graph_id in v["meta"]["graphs"]
        ]
        return np.unique(node_types)

    def node_attributes(self, graph_id, node_type):
        """Return a list of node attribute names, for nodes of node_type type, that belong to graph with graph_id"""
        x_ind = []
        nodes = [
            v
            for v in self.G["vertices"]
            if (graph_id in v["meta"]["graphs"]) and (node_type in v["meta"]["label"])
        ]
        for v in nodes:
            for k in list(v["data"].keys()):
                x_ind.append(k)

        return np.unique(x_ind)

    def node_attr_dim(self, graph_id, node_type):
        """Return the dimensionality of node attributes, for nodes of node_type type, that belong to graph with graph_id"""
        return len(np.unique(self.node_attributes(graph_id, node_type)))

    def __init__(self, G, node_attributes=None, node_labels=None):
        if "networkx.classes.graph.Graph" in str(G.__class__):
            G_json = self._nx_to_json(G)
            self.G = self._json_to_epgm(G_json, node_attributes, node_labels)
        elif "str" in str(G.__class__):  # this assumes that G is the path to EPGM graph
            self.G = self.load(G)
        else:
            raise Exception("G has unknown class ", str(G.__class__))

        self.G_nx = {}  # placeholder for storing graphs in networkx format (populated by calling the .to_nx() method)

    def append(self, G_add):
        """Update self.G by adding a new graph G_json_add"""
        G_json_add = self._nx_to_json(G_add)
        if any([G_json_add["graph"]["id"] in g["id"] for g in self.G["graphs"]]):
            raise Exception(
                "Graph with id {} already exists".format(G_json_add["graph"]["id"])
            )

        # update self.G['graphs']:
        G_epgm_add = self._json_to_epgm(G_json_add, None, None)  # was None, None, None
        self.G["graphs"].append(G_epgm_add["graphs"][0])

        # update self.G['vertices']:
        G_vertices = set([v["id"] for v in self.G["vertices"]])
        G_add_vertices = set([v["id"] for v in G_epgm_add["vertices"]])
        common_vertices = list(G_vertices.intersection(G_add_vertices))
        new_vertices = list(G_add_vertices.difference(G_vertices))

        # update common vertices in self.G and G_epgm_add:
        for v_id in common_vertices:
            i = [v["id"] == v_id for v in self.G["vertices"]].index(
                True
            )  # find the vertex index in self.G
            self.G["vertices"][i]["meta"]["graphs"].append(
                G_epgm_add["graphs"][0]["id"]
            )

        # append new vertices to self.G['vertices']:
        for v_id in new_vertices:
            i = [v["id"] == v_id for v in G_epgm_add["vertices"]].index(
                True
            )  # find the index of v_id in G_epgm_add['vertices']
            self.G["vertices"].append(
                G_epgm_add["vertices"][i]
            )  # add the vertex v_id to self.G['vertices']

        # update self.G['edges']:
        G_edges = set([(e["source"], e["target"]) for e in self.G["edges"]])
        G_add_edges = set([(e["source"], e["target"]) for e in G_epgm_add["edges"]])
        common_edges = list(G_edges.intersection(G_add_edges))
        new_edges = list(G_add_edges.difference(G_edges))

        # update common edges in self.G and G_epgm_add:
        for e_id in common_edges:
            i = [(e["source"], e["target"]) == e_id for e in self.G["edges"]].index(
                True
            )  # find the edge index in self.G
            self.G["edges"][i]["meta"]["graphs"].append(G_epgm_add["graphs"][0]["id"])

        # append new edges to self.G['edges']:
        for e_id in new_edges:
            i = [(e["source"], e["target"]) == e_id for e in G_epgm_add["edges"]].index(
                True
            )  # find the index of e_id in G_epgm_add['edges']
            self.G["edges"].append(
                G_epgm_add["edges"][i]
            )  # add the vertex e_id to self.G['edges']

    def to_nx_OLD(
        self,
        graph_id,
        directed=False,
        parallel_processing=True,
        n_jobs=multiprocessing.cpu_count(),
        progress=True,
        chunksize=100,
    ):
        """Convert the graph specified by its graph_id to networkx graph"""

        if (
            graph_id in self.G_nx.keys()
        ):  # if self.G_nx[graph_id] already exists, just return it, otherwise evaluate it
            return self.G_nx[graph_id]
        else:
            print("Converting the EPGM graph {} to NetworkX graph...".format(graph_id))
            if not any([graph_id in g["id"] for g in self.G["graphs"]]):
                raise Exception("Graph with id {} does not exist".format(graph_id))

            # List relevant nodes and edges:
            print("...extracting relevant nodes...", end="")
            nodes = [
                v["id"] for v in self.G["vertices"] if graph_id in v["meta"]["graphs"]
            ]
            print(" ...{} nodes extracted...".format(len(nodes)))
            print("...extracting relevant edges...", end="")
            edges = [
                (e["source"], e["target"])
                for e in self.G["edges"]
                if graph_id in e["meta"]["graphs"]
            ]
            print(" ...{} edges extracted...".format(len(edges)))
            # TODO: implement the case of weighted edges

            # create a graph as dict of lists in the format (node_id: [neighbour nodes])
            print("...building the graph as dict of lists...")
            print(
                "...[parallel_processing: {}, n_jobs: {}, progress_bar: {}]".format(
                    parallel_processing, n_jobs, progress
                )
            )

            if parallel_processing:  # parallel execution
                pool = Pool(processes=n_jobs)
                if progress:
                    n = len(nodes)
                    self.G_nx[graph_id] = []

                    # pbar = ProgressBar(
                    #     widgets=[
                    #         SimpleProgress(
                    #             format="%(value_s)s of %(max_value_s)s nodes processed (%(percentage)3d%%)"
                    #         )
                    #     ],
                    #     maxval=n,
                    # ).start()
                    # _ = [pool.apply_async(partial(node_neighbours, edges=edges), args=(v,),
                    #                       callback=self.G_nx[graph_id].append) for v in nodes]
                    # it seems that appending results using callback works much slower than either pool.map_async or pool.map
                    # while len(self.G_nx[graph_id]) != n:
                    #     pbar.update(len(self.G_nx[graph_id]))
                    #     sleep(1)

                    graph = pool.imap(
                        partial(node_neighbours, edges=edges), nodes, chunksize
                    )  # lazy map
                    # evaluate batches of imap, as the progress bar is being updated:
                    while len(self.G_nx[graph_id]) != n:
                        self.G_nx[graph_id].append(next(graph))
                        # pbar.update(len(self.G_nx[graph_id]))

                    # pbar.finish()

                    self.G_nx[graph_id] = dict(self.G_nx[graph_id])

                else:
                    self.G_nx[graph_id] = dict(
                        pool.map(partial(node_neighbours, edges=edges), nodes)
                    )

                pool.close()
                pool.join()

            else:  # sequential execution
                self.G_nx[graph_id] = {
                    v: [e[1] for e in edges if e[0] == v] for v in nodes
                }  # this works ~2.5x faster (for cora dataset) than the above for loop

            print("...converting the graph to nx format...")
            self.G_nx[graph_id] = nx.from_dict_of_lists(self.G_nx[graph_id])

            if directed:
                self.G_nx[graph_id] = self.G_nx[graph_id].to_directed()
            else:
                self.G_nx[graph_id] = self.G_nx[graph_id].to_undirected()

            return self.G_nx[graph_id]

    def to_nx(self, graph_id, directed=False, *args):
        """Convert the graph specified by its graph_id to networkx Directed Multi-graph"""

        if (
            False
        ):  # graph_id in self.G_nx.keys():  # if self.G_nx[graph_id] already exists, just return it, otherwise evaluate it
            return self.G_nx[graph_id]
        else:  # we always re-calculate self.G_nx, since directed argument can change
            print("Converting the EPGM graph {} to NetworkX graph...".format(graph_id))
            if not any([graph_id in g["id"] for g in self.G["graphs"]]):
                raise Exception("Graph with id {} does not exist".format(graph_id))

        self.G_nx[
            graph_id
        ] = (
            nx.MultiDiGraph()
        )  # create an empty directed graph that can store multiedges
        # add nodes to self.G_nx[graph_id], together with their attributes stored in 'data':
        self.G_nx[graph_id].add_nodes_from(
            [
                (v["id"], {**v["data"], **{"label": v["meta"].get("label", "")}})
                for v in self.G["vertices"]
            ]
        )
        # add edges to self.G_nx[graph_id], together with their attributes stored in 'data':
        # I have added the edge label in the edge data; sets the label to '' if the edges don't have a label
        self.G_nx[graph_id].add_edges_from(
            [
                (
                    e["source"],
                    e["target"],
                    e["id"],
                    {**e["data"], **{"label": e["meta"].get("label", "")}},
                )
                for e in self.G["edges"]
            ]
        )

        if not directed:
            self.G_nx[graph_id] = self.G_nx[graph_id].to_undirected()

        return self.G_nx[graph_id]

    def adjacency(self, graph_id, directed=False):
        """Return adjacency matrix of a graph specified by its graph_id"""

        if not any([graph_id in g["id"] for g in self.G["graphs"]]):
            raise Exception("Graph with id {} does not exist".format(graph_id))

        print("...building the adjacency matrix...")
        nodes = [v["id"] for v in self.G["vertices"] if graph_id in v["meta"]["graphs"]]
        adj = nx.adjacency_matrix(
            self.to_nx(graph_id, directed), nodelist=nodes
        )  # ensure the nodes in adj are ordered the same as in the epgm graph

        return adj

    # def adjacency_sans_nx(self, graph_id, directed=False):
    #     """Compose the adjacency matrix of a graph specified by its graph_id, NOT using networkx"""
    #     if not any([graph_id in g['id'] for g in self.G['graphs']]):
    #         raise Exception("Graph with id {} does not exist".format(graph_id))
    #
    #     # List relevant nodes and edges:
    #     nodes = [v['id'] for v in self.G['vertices'] if graph_id in v['meta']['graphs']]
    #     edges = [(e['source'], e['target']) for e in self.G['edges'] if graph_id in e['meta']['graphs']]
    #
    #     n_nodes = len(nodes)
    #     adj = sp.lil_matrix((n_nodes, n_nodes), dtype=int)  # TODO: implement the case of weighted edges
    #     for i, v in enumerate(nodes):
    #         data = data = partial(node_neighbours_extended, nodes=nodes, edges=edges)(v) # node_neighbours_extended(v, nodes, edges)
    #         nbr_idx = data[1][0]
    #         adj.rows[i] = list(nbr_idx)
    #         adj.data[i] = [1] * len(nbr_idx)
    #         # TODO: implement the case of weighted edges
    #
    #     if not directed:  # symmetrize the adj matrix
    #         adj = adj + adj.T - sp.diags(adj.diagonal(), dtype=int)
    #
    #     return adj

    # def adjacency_from_edgelist(self, graph_id, directed=False):
    #     """Return adjacency matrix of a graph specified by its graph_id"""
    #     # FIXME: the order of nodes in the created graph, and hence in adj, is different from that in the EPGM graph. Do not use this unless it's fixed!
    #     if not any([graph_id in g['id'] for g in self.G['graphs']]):
    #         raise Exception("Graph with id {} does not exist".format(graph_id))
    #
    #     # An alternative way, works faster, BUT GIVES A DIFFERENT NODES ORDER!!!
    #     # List relevant edges:
    #     edges = [(int(e['source']), int(e['target'])) for e in self.G['edges'] if graph_id in e['meta']['graphs']]
    #     # TODO: implement the case of weighted edges
    #     graph = nx.from_edgelist(edges)
    #     if directed:
    #         graph = graph.to_directed()
    #     else:
    #         graph = graph.to_undirected()
    #
    #     adj = nx.adjacency_matrix(graph)
    #
    #     return adj

    def edgelist(self, graph_id, directed=False):
        """Return edgelist of a graph specified by its graph_id"""

        print("...extracting the edgelist...")
        # edgelist = nx.to_edgelist(self.to_nx(graph_id, directed))
        edgelist = [
            (e["source"], e["target"])
            for e in self.G["edges"]
            if graph_id in e["meta"]["graphs"]
        ]  # works much faster, gets the edgelist directly from the edges.json part of epgm graph
        print(" ...{} edges extracted...".format(len(edgelist)))

        return edgelist

    def save(self, path):
        """
        Write self.G into three json files: graphs.json, vertices.json, and edges.json, in path directory
        """
        self.path = path
        if not os.path.isdir(self.path):
            os.makedirs(self.path)

        for k in self.G.keys():
            with open(os.path.join(self.path, str(k) + ".json"), "w") as fp:
                if isinstance(self.G[k], list):
                    for l in self.G[k]:
                        json.dump(l, fp)
                        fp.write("\n")
                elif isinstance(self.G[k], OrderedDict) or isinstance(self.G[k], dict):
                    json.dump(self.G[k], fp)
                else:
                    print(
                        "type(G[",
                        k,
                        "]): unexpected type",
                        type(self.G[k]),
                        ", stopping.",
                    )
                    raise ()

    def save_as_graphml(self, graph_id, fname, directed):
        """
        Save the graph in GraphML XML format (e.g., for visualisation in gephi)
        Args:
            graph_id: unique id of the graph
            fname: file name to save the graphml into

        Returns:
            Exit code of nx.write_graphml()

        """
        if not any([graph_id in g["id"] for g in self.G["graphs"]]):
            raise Exception("Graph with id {} does not exist".format(graph_id))

        return nx.write_graphml(nx.DiGraph(self.to_nx(graph_id, directed)), path=fname)



# -*- coding: utf-8 -*-
#
# Copyright 2018 Data61, CSIRO
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mappers to provide input data for the graph models in layers.

"""
__all__ = ["NodeSequence", "GraphSAGENodeGenerator", "HinSAGENodeGenerator"]

import operator
from functools import reduce

import numpy as np
import itertools as it
from keras.utils import Sequence

from ..data.explorer import (
    SampledBreadthFirstWalk,
    SampledHeterogeneousBreadthFirstWalk,
)
from ..core.graph import StellarGraphBase
from ..core.utils import is_real_iterable


class NodeSequence(Sequence):
    """Keras-compatible data generator to use with the Keras
    methods :meth:`keras.Model.fit_generator`, :meth:`keras.Model.evaluate_generator`,
    and :meth:`keras.Model.predict_generator`.

    This class generated data samples for node inference models
    and should be created using the `.flow(...)` method of
    :class:`GraphSAGENodeGenerator` or :class:`HinSAGENodeGenerator`.

    These Generators are classes that capture the graph structure
    and the feature vectors of each node. These generator classes
    are used within the NodeSequence to generate samples of k-hop
    neighbourhoods in the graph and to return to this class the
    features from the sampled neighbourhoods.

    Args:
        generator: GraphSAGENodeGenerator or HinSAGENodeGenerator
            The generator object containing the graph information.
        ids: list
            A list of the node_ids to be used as head-nodes in the
            downstream task.
        targets: list, optional (default=None)
            A list of targets or labels to be used in the downstream
            class.
    """

    def __init__(self, generator, ids, targets=None):
        # Check that ids is an iterable
        if not is_real_iterable(ids):
            raise TypeError("IDs must be an iterable or numpy array of graph node IDs")

        # Check targets is iterable & has the correct length
        if targets is not None:
            if not is_real_iterable(targets):
                raise TypeError("Targets must be None or an iterable or numpy array ")
            if len(ids) != len(targets):
                raise ValueError(
                    "The length of the targets must be the same as the length of the ids"
                )

        # Infer head_node_type
        # TODO: Generalize to multiple head node target types?
        head_node_types = set([generator.schema.get_node_type(n) for n in ids])
        if len(head_node_types) > 1:
            raise ValueError(
                "Only a single head node type is currently supported for HinSAGE models"
            )
        head_node_type = head_node_types.pop()

        # Store the generator to draw samples from graph
        self.generator = generator
        self.ids = list(ids)
        self.targets = targets
        self.data_size = len(self.ids)

        # Save head node type and generate sampling schema
        self.head_node_types = [head_node_type]
        self._sampling_schema = generator.schema.sampling_layout(
            self.head_node_types, generator.num_samples
        )

    def __len__(self):
        "Denotes the number of batches per epoch"
        return int(np.ceil(self.data_size / self.generator.batch_size))

    def __getitem__(self, batch_num):
        """
        Generate one batch of data

        Args:
            batch_num (int): number of a batch

        Returns:
            batch_feats (list): Node features for nodes and neighbours sampled from a
                batch of the supplied IDs
            batch_targets (list): Targets/labels for the batch.

        """
        start_idx = self.generator.batch_size * batch_num
        end_idx = start_idx + self.generator.batch_size

        if start_idx >= self.data_size:
            raise IndexError("Mapper: batch_num larger than length of data")
        # print("Fetching batch {} [{}]".format(batch_num, start_idx))

        # Get head nodes
        head_ids = self.ids[start_idx:end_idx]

        # Get targets for nodes
        if self.targets is None:
            batch_targets = None
        else:
            batch_targets = self.targets[start_idx:end_idx]

        # Get sampled nodes
        batch_feats = self.generator.sample_features(head_ids, self._sampling_schema)

        return batch_feats, batch_targets


class GraphSAGENodeGenerator:
    """A data generator for node prediction with Homogeneous GraphSAGE models

    At minimum, supply the StellarGraph, the batch size, and the number of
    node samples for each layer of the GraphSAGE model.

    The supplied graph should be a StellarGraph object that is ready for
    machine learning. Currently the model requires node features for all
    nodes in the graph.

    Use the :meth:`flow` method supplying the nodes and (optionally) targets
    to get an object that can be used as a Keras data generator.

    Example::

        G_generator = GraphSAGENodeGenerator(G, 50, [10,10])
        train_data_gen = G_generator.flow(node_ids)

    Args:
        G (StellarGraph): The machine-learning ready graph.
        batch_size (int): Size of batch to return.
        num_samples (list): The number of samples per layer (hop) to take.
        seed (int): [Optional] Random seed for the node sampler.
        name (str or None): Name of the generator (optional)
    """

    def __init__(self, G, batch_size, num_samples, seed=None, name=None):
        if not isinstance(G, StellarGraphBase):
            raise TypeError("Graph must be a StellarGraph object.")

        self.graph = G
        self.num_samples = num_samples
        self.batch_size = batch_size
        self.name = name

        # Check if the graph has features
        G.check_graph_for_ml()

        # Create sampler for GraphSAGE
        self.sampler = SampledBreadthFirstWalk(G, seed=seed)

        # We need a schema for compatibility with HinSAGE
        self.schema = G.create_graph_schema(create_type_maps=True)

        # Check that there is only a single node type for GraphSAGE
        if len(self.schema.node_types) > 1:
            print(
                "Warning: running homogeneous GraphSAGE on a graph with multiple node types"
            )

    def sample_features(self, head_nodes, sampling_schema):
        """
        Sample neighbours recursively from the head nodes, collect the features of the
        sampled nodes, and return these as a list of feature arrays for the GraphSAGE
        algorithm.

        Args:
            head_nodes: An iterable of head nodes to perform sampling on.
            sampling_schema: The sampling schema for the model

        Returns:
            A list of the same length as ``num_samples`` of collected features from
            the sampled nodes of shape:
            ``(len(head_nodes), num_sampled_at_layer, feature_size)``
            where num_sampled_at_layer is the cumulative product of `num_samples`
            for that layer.
        """
        node_samples = self.sampler.run(nodes=head_nodes, n=1, n_size=self.num_samples)

        # Reshape node samples to sensible format
        def get_levels(loc, lsize, samples_per_hop, walks):
            end_loc = loc + lsize
            walks_at_level = list(it.chain(*[w[loc:end_loc] for w in walks]))
            if len(samples_per_hop) < 1:
                return [walks_at_level]
            return [walks_at_level] + get_levels(
                end_loc, lsize * samples_per_hop[0], samples_per_hop[1:], walks
            )

        nodes_per_hop = get_levels(0, 1, self.num_samples, node_samples)
        node_type = sampling_schema[0][0][0]

        # Get features for sampled nodes
        batch_feats = [
            self.graph.get_feature_for_nodes(layer_nodes, node_type)
            for layer_nodes in nodes_per_hop
        ]

        # Resize features to (batch_size, n_neighbours, feature_size)
        batch_feats = [
            np.reshape(a, (len(head_nodes), -1 if np.size(a) > 0 else 0, a.shape[1]))
            for a in batch_feats
        ]
        return batch_feats

    def flow(self, node_ids, targets=None):
        """
        Creates a generator/sequence object for training or evaluation
        with the supplied node ids and numeric targets.

        The node IDs are the nodes to train or inference on: the embeddings
        calculated for these nodes are passed to the downstream task. These
        are a subset of the nodes in the graph.

        The targets are an array of numeric targets corresponding to the
        supplied node_ids to be used by the downstream task. They should
        be given in the same order as the list of node IDs.
        If they are not specified (for example, for use in prediction),
        the targets will not be available to the downsteam task.

        Args:
            node_ids: an iterable of node IDs
            targets: a 2D array of numeric targets with shape
                `(len(node_ids), target_size)`

        Returns:
            A NodeSequence object to use with the GraphSAGE model
            in Keras methods ``fit_generator``, ``evaluate_generator``,
            and ``predict_generator``

        """
        return NodeSequence(self, node_ids, targets)

    def flow_from_dataframe(self, node_targets):
        """
        Creates a generator/sequence object for training or evaluation
        with the supplied node ids and numeric targets.

        Args:
            node_targets: a Pandas DataFrame of numeric targets indexed
                by the node ID for that target.

        Returns:
            A NodeSequence object to use with the GraphSAGE model
            in Keras methods ``fit_generator``, ``evaluate_generator``,
            and ``predict_generator``

        """
        return NodeSequence(self, node_targets.index, node_targets.values)


class HinSAGENodeGenerator:
    """Keras-compatible data mapper for Heterogeneour GraphSAGE (HinSAGE)

     At minimum, supply the StellarGraph, the batch size, and the number of
     node samples for each layer of the HinSAGE model.

     The supplied graph should be a StellarGraph object that is ready for
     machine learning. Currently the model requires node features for all
     nodes in the graph.

     Use the :meth:`flow` method supplying the nodes and (optionally) targets
     to get an object that can be used as a Keras data generator.

     Example::

         G_generator = HinSAGENodeGenerator(G, 50, [10,10])
         data_gen = G_generator.flow(node_ids)

    Args:
        G (StellarGraph): The machine-learning ready graph
        batch_size (int): Size of batch to return
        num_samples (list): The number of samples per layer (hop) to take
        seed (int), Optional: Random seed for the node sampler
        name (str), optional: Name of the generator.
     """

    def __init__(self, G, batch_size, num_samples, seed=None, name=None):

        self.graph = G
        self.num_samples = num_samples
        self.batch_size = batch_size
        self.name = name

        # We require a StellarGraph
        if not isinstance(G, StellarGraphBase):
            raise TypeError("Graph must be a StellarGraph object.")

        G.check_graph_for_ml(features=True)

        # Create sampler for GraphSAGE
        self.sampler = SampledHeterogeneousBreadthFirstWalk(G, seed=seed)

        # Generate schema
        self.schema = G.create_graph_schema(create_type_maps=True)

    def sample_features(self, head_nodes, sampling_schema):
        """
        Sample neighbours recursively from the head nodes, collect the features of the
        sampled nodes, and return these as a list of feature arrays for the GraphSAGE
        algorithm.

        Args:
            head_nodes: An iterable of head nodes to perform sampling on.
            node_sampling_schema: The sampling schema for the HinSAGE model,
                this is can be generated by the ``GraphSchema`` object.

        Returns:
            A list of the same length as ``num_samples`` of collected features from
            the sampled nodes of shape:
            ``(len(head_nodes), num_sampled_at_layer, feature_size)``
            where num_sampled_at_layer is the cumulative product of `num_samples`
            for that layer.
        """
        # Get sampled nodes
        node_samples = self.sampler.run(nodes=head_nodes, n=1, n_size=self.num_samples)

        # Reshape node samples to the required format for the HinSAGE model
        # This requires grouping the sampled nodes by edge type and in order
        nodes_by_type = [
            (
                nt,
                reduce(
                    operator.concat,
                    (samples[ks] for samples in node_samples for ks in indices),
                ),
            )
            for nt, indices in sampling_schema[0]
        ]

        # Get features
        batch_feats = [
            self.graph.get_feature_for_nodes(layer_nodes, nt)
            for nt, layer_nodes in nodes_by_type
        ]

        # Resize features to (batch_size, n_neighbours, feature_size)
        batch_feats = [
            np.reshape(a, (len(head_nodes), -1 if np.size(a) > 0 else 0, a.shape[1]))
            for a in batch_feats
        ]

        return batch_feats

    def flow(self, node_ids, targets=None):
        """
        Creates a generator/sequence object for training or evaluation
        with the supplied node ids and numeric targets.

        The node IDs are the nodes to train or inference on: the embeddings
        calculated for these nodes are passed to the downstream task. These
        are a subset of the nodes in the graph.

        The targets are an array of numeric targets corresponding to the
        supplied node_ids to be used by the downstream task. They should
        be given in the same order as the list of node IDs.
        If they are not specified (for example, for use in prediction),
        the targets will not be available to the downsteam task.

        Args:
            node_ids (iterable): The head node IDs
            targets (Numpy array): a 2D array of numeric targets with shape
                ``(len(node_ids), target_size)``
            node_type (str), optional: The target node type, if not given
                the node type will be inferred from the graph.

        Returns:
            A NodeSequence object to use with the GraphSAGE model
            in Keras methods `fit_generator`, `evaluate_generator`,
            and `predict_generator`.

        """
        return NodeSequence(self, node_ids, targets)

    def flow_from_dataframe(self, node_targets):
        """
        Creates a generator/sequence object for training or evaluation
        with the supplied node ids and numeric targets.

        Args:
            node_targets (DataFrame): Numeric targets indexed
                by the node ID for that target.
            node_type (str), optional: The target node type, if not given
                the node type will be inferred from the graph.

        Returns:
            A NodeSequence object to use with the GraphSAGE model
            in Keras methods `fit_generator`, `evaluate_generator`,
            and `predict_generator`.
        """

        return NodeSequence(self, node_targets.index, node_targets.values)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2018 Data61, CSIRO
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The StellarGraph class that encapsulates information required for
a machine-learning ready graph used by models.

"""
__all__ = ["StellarGraph", "StellarDiGraph", "StellarGraphBase"]

from stellargraph.core.schema import EdgeType

import random
import itertools as it

import pandas as pd
import numpy as np
from networkx.classes.multigraph import MultiGraph
from networkx.classes.multidigraph import MultiDiGraph

from collections import Iterable, Iterator

from .. import globalvar
from .schema import GraphSchema
from .utils import is_real_iterable


def _convert_from_node_attribute(
    G, attr_name, node_types, node_type_name=None, dtype="f"
):
    """
    Transform the node attributes to feature vectors, for use with machine learning models.

    Each node is assumed to have a numeric array stored in the attribute_name and
    which is suitable for use in machine learning models.

    Args:
        G: NetworkX graph
        attr_name: Name of node attribute to use for conversion
        node_types: Node types in graph
        node_type_name: (optional) The name of the node attribute specifying the type.
        dtype: (optional) The numpy datatype to create the features array.

    Returns:
        index_map: a dictionary of node_type -> {node_id: node_index}
        attribute_arrays: a dictionary of node_type -> numpy array storing the features
    """
    attribute_arrays = {}
    node_index_map = {}

    # Enumerate all nodes in graph
    nodes_by_type = {
        nt: [n for n, ndata in G.nodes(data=True) if ndata[node_type_name] == nt]
        for nt in node_types
    }

    # Get the target values for each node type
    for nt in node_types:
        nt_node_list = nodes_by_type[nt]

        # Add None to node list as ID of unknown nodes
        nt_node_list.append(None)

        # Create map between node id and index (including None)
        node_index_map[nt] = {nid: ii for ii, nid in enumerate(nt_node_list)}

        # The node data
        attr_data = [v if v is None else G.node[v].get(attr_name) for v in nt_node_list]

        # Get the size of the features
        data_sizes = {
            np.size(G.node[v].get(attr_name, [])) for v in nt_node_list if v is not None
        }

        # Warn if nodes don't have the attribute
        if 0 in data_sizes:
            print(
                "Warning: Some nodes have no value for attribute '{}', "
                "using default value.".format(attr_name)
            )
            data_sizes.discard(0)

        # Check all are the same for this node type
        if len(data_sizes) > 1:
            raise ValueError(
                "Data sizes in nodes of type {} are inconsistent "
                "for the attribute '{}' ".format(nt, attr_name)
            )

        # If some node_type have no nodes with the attribute, skip them
        if len(data_sizes) == 0:
            continue

        # Create zero attribute array
        data_size = data_sizes.pop()

        # Dummy feature/target value for invalid nodes,
        # this will be inserted into the array in two cases:
        # 1. node ID of None (representing sampling for a missing neighbour)
        # 2. node with no attribute
        # TODO: Make these two cases more explicit, allow custom values.
        default_value = np.zeros(data_size)

        # Convert to numpy array
        attribute_arrays[nt] = np.asarray(
            [x if x is not None else default_value for x in attr_data]
        )

    return node_index_map, attribute_arrays


def _convert_from_node_data(data, node_type_map, node_types, dtype="f"):
    """
    Store the node data as feature vectors, for use with machine learning models.

    For a single node type, the data can be either:
     * a Pandas DataFrame with the index being node IDs and the columns the numeric
        feature values. Note that the features must be numeric.
     * a list or iterable of `(node_id, node_feature)` pairs where node_feature is
        a value, a list of values, or a numpy array representing the numeric feature
        values.

    For multiple node types, the data can be either:
     * a dictionary of node_type -> DataFrame with the index of each DataFrame
        being node IDs and the columns the numeric feature values.
        Note that the features must be numeric and can be different sizes for each
        node type.
     * a list or iterable of `(node_id, node_feature)` pairs where node_feature is
        a value, a list of values, or a numpy array representing the numeric feature
        values.

    Args:
        data: dict, list or DataFrame
            The data for the nodes, partitioned by node type

        node_type_map: dict
            Mapping of node_id to node_type

        node_types: list
            List of the node types in the data

        dtype: Numpy datatype optional (default='float32')
            The numpy datatype to create the features array.

    Returns:
        index_map: a dictionary of node_type -> {node_id: node_index}
        attribute_arrays: a dictionary of node_type -> numpy array storing the features
    """
    # if data is a dict of pandas dataframes or iterators, pull the features for each node type in the dictionary
    if isinstance(data, dict):
        # The keys should match the node types
        if not all(k in node_types for k in data.keys()):
            raise ValueError(
                "All node types in supplied feature dict should be in the graph"
            )

        data_arrays = {}
        data_index = {}
        for nt, arr in data.items():
            if isinstance(arr, pd.DataFrame):
                node_index_map = {nid: nii for nii, nid in enumerate(arr.index)}
                try:
                    data_arr = arr.values.astype(dtype)
                except ValueError:
                    raise ValueError(
                        "Node data passed as Pandas arrays should contain only numeric values"
                    )

            elif isinstance(arr, (Iterable, list)):
                data_arr = []
                node_index_map = {}
                for ii, (node_id, datum) in enumerate(arr):
                    data_arr.append(datum)
                    node_index_map[node_id] = ii
                data_arr = np.vstack(data_arr)

            else:
                raise TypeError(
                    "Node data should be a pandas array, an iterable, a list, or name of a node_attribute"
                )

            # Add default value to end of feature array
            default_value = np.zeros(data_arr.shape[1])
            data_arrays[nt] = np.vstack([data_arr, default_value])

            node_index_map[None] = data_arr.shape[0]
            data_index[nt] = node_index_map

    # If data is a pd.Dataframe, try pulling out the type
    elif isinstance(data, pd.DataFrame):
        if len(node_types) > 1:
            raise TypeError(
                "When there is more than one node type, pass node features as a dictionary."
            )
        node_type = next(iter(node_types))
        data_index, data_arrays = _convert_from_node_data(
            {node_type: data}, node_type_map, node_types, dtype
        )

    # If data an iterator try recreating the nodes by type
    elif isinstance(data, (Iterator, list)):
        node_data_by_type = {nt: [] for nt in node_types}
        for d in data:
            node_type = node_type_map.get(d[0])
            if node_type is None:
                raise TypeError("Node type not found in importing feature vectors!")

            node_data_by_type[node_type].append(d)

        data_index, data_arrays = _convert_from_node_data(
            node_data_by_type, node_type_map, node_types, dtype
        )

    else:
        raise TypeError(
            "Node data should be a dictionary, a pandas array, an iterable, or a tuple."
        )

    return data_index, data_arrays


class StellarGraphBase:
    """
    StellarGraph class for undirected graph ML models. It stores both
    graph information from a NetworkX Graph object as well as features
    for machine learning.

    To create a StellarGraph object ready for machine learning, at a
    minimum pass the graph structure to the StellarGraph as a NetworkX
    graph:

    For undirected models::

        Gs = StellarGraph(nx_graph)


    For directed models::

        Gs = StellarDiGraph(nx_graph)


    To create a StellarGraph object with node features, supply the features
    as a numeric feature vector for each node.

    To take the feature vectors from a node attribute in the original NetworkX
    graph, supply the attribute name to the ``node_features`` argument::

        Gs = StellarGraph(nx_graph, node_features="feature")


    where the nx_graph contains nodes that have a "feature" attribute containing
    the feature vector for the node. All nodes of the same type must have
    the same size feature vectors.

    Alternatively, supply the node features as Pandas DataFrame objects with
    the of the DataFrame set to the node IDs. For graphs with a single node
    type, you can supply the DataFrame object directly to StellarGraph::

        node_data = pd.DataFrame(
            [feature_vector_1, feature_vector_2, ..],
            index=[node_id_1, node_id_2, ...])
        Gs = StellarGraph(nx_graph, node_features=node_data)

    For graphs with multiple node types, provide the node features as Pandas
    DataFrames for each type separately, as a dictionary by node type.
    This allows node features to have different sizes for each node type::

        node_data = {
            node_type_1: pd.DataFrame(...),
            node_type_2: pd.DataFrame(...),
        }
        Gs = StellarGraph(nx_graph, node_features=node_data)


    You can also supply the node feature vectors as an iterator of `node_id`
    and feature vector pairs, for graphs with single and multiple node types::

        node_data = zip([node_id_1, node_id_2, ...],
            [feature_vector_1, feature_vector_2, ..])
        Gs = StellarGraph(nx_graph, node_features=node_data)


    Args:
        node_type_name: str, optional (default=globals.TYPE_ATTR_NAME)
            This is the name for the node types that StellarGraph uses
            when processing heterogeneous graphs. StellarGraph will
            look for this attribute in the nodes of the graph to determine
            their type.

        edge_type_name: str, optional (default=globals.TYPE_ATTR_NAME)
            This is the name for the edge types that StellarGraph uses
            when processing heterogeneous graphs. StellarGraph will
            look for this attribute in the edges of the graph to determine
            their type.

        node_features: str, dict, list or DataFrame optional (default=None)
            This tells StellarGraph where to find the node feature information
            required by some graph models. These are expected to be
            a numeric feature vector for each node in the graph.

    """

    def __init__(self, incoming_graph_data=None, **attr):
        # TODO: add doc string
        super().__init__(incoming_graph_data, **attr)

        # Names of attributes that store the type of nodes and edges
        self._node_type_attr = attr.get("node_type_name", globalvar.TYPE_ATTR_NAME)
        self._edge_type_attr = attr.get("edge_type_name", globalvar.TYPE_ATTR_NAME)

        # Names for the feature/target type (used if they are supplied and
        #  feature/target spec not supplied"
        self._feature_attr = attr.get("feature_name", globalvar.FEATURE_ATTR_NAME)
        self._target_attr = attr.get("target_name", globalvar.TARGET_ATTR_NAME)

        # Ensure that the incoming graph data has node & edge types
        # TODO: This requires traversing all nodes and edges. Is there another way?
        # TODO: Should we add the default values as class arguments for these?
        node_types = set()
        type_for_node = {}
        for n, ndata in self.nodes(data=True):
            if self._node_type_attr not in ndata:
                ndata[self._node_type_attr] = "default"
            type_for_node[n] = ndata[self._node_type_attr]
            node_types.add(ndata[self._node_type_attr])

        edge_types = set()
        for n1, n2, k, edata in self.edges(keys=True, data=True):
            if self._edge_type_attr not in edata:
                edata[self._edge_type_attr] = "default"
            edge_types.add(edata[self._edge_type_attr])

        # New style: we are passed numpy arrays or pandas arrays of the feature vectors
        node_features = attr.get("node_features", None)
        dtype = attr.get("dtype", "float32")

        # If node_features is a string, load features from this attribute of the nodes in the graph
        if isinstance(node_features, str):
            data_index_maps, data_arrays = _convert_from_node_attribute(
                self, node_features, node_types, self._node_type_attr, dtype
            )

        # Otherwise try importing node_features as a Numpy array or Pandas Dataframe
        elif node_features is not None:
            data_index_maps, data_arrays = _convert_from_node_data(
                node_features, type_for_node, node_types, dtype
            )

        else:
            data_index_maps = {}
            data_arrays = {}

        # TODO: What other convenience attributes do we need?
        self._nodes_by_type = None

        # This stores the feature vectors per node type as numpy arrays
        self._node_attribute_arrays = data_arrays

        # This stores the map between node ID and index in the attribute arrays
        self._node_index_maps = data_index_maps

    def __repr__(self):
        directed_str = "Directed" if self.is_directed() else "Undirected"
        s = "{}: {} multigraph\n".format(type(self).__name__, directed_str)
        s += "    Nodes: {}, Edges: {}\n".format(
            self.number_of_nodes(), self.number_of_edges()
        )
        return s

    def fit_attribute_spec(self, *args, **kwargs):
        print("Fit attribute spec is deprecated")
        pass

    def check_graph_for_ml(self, features=True):
        """
        Checks if all properties required for machine learning training/inference are set up.
        An error will be raised if the graph is not correctly setup.
        """
        # TODO: This are simple tests and miss many problems that could arise, improve!
        # TODO: At this point perhaps we should generate features rather than in fit_attribute_spec
        # TODO: but if so how do we know whether to fit the attribute specs or not?
        # Check features on the nodes:
        if features and len(self._node_attribute_arrays) == 0:
            raise RuntimeError(
                "This StellarGraph has no numeric feature attributes for nodes"
                "Node features are required for machine learning"
            )

        # How about checking the schema?

    def get_feature_for_nodes(self, nodes, node_type=None):
        """
        Get the numeric feature vectors for the specified node or nodes.
        If the node type is not specified the node types will be found
        for all nodes. It is therefore important to supply the ``node_type``
        for this method to be fast.

        Args:
            n: (list or hashable) Node ID or list of node IDs
            node_type: (hashable) the type of the nodes.

        Returns:
            Numpy array containing the node features for the requested nodes.
        """
        # TODO: change method's name to node_features(), and add @property decorator
        if not is_real_iterable(nodes):
            nodes = [nodes]

        # Get the node type
        # TODO: This is slow, refactor so that self._node_index_maps gives the node type
        if node_type is None:
            node_types = {
                self.node[n].get(self._node_type_attr) for n in nodes if n is not None
            }

            if None in node_types:
                raise ValueError(
                    "All nodes must have a type specified as the "
                    "'{}' attribute.".format(self._node_type_attr)
                )

            if len(node_types) > 1:
                raise ValueError("All nodes must be of the same type.")

            if len(node_types) == 0:
                raise ValueError(
                    "At least one node must be given if node_type not specified"
                )

            node_type = node_types.pop()

        # Check node_types
        if (
            node_type not in self._node_attribute_arrays
            or node_type not in self._node_index_maps
        ):
            raise ValueError("Features not found for node type '{}'")

        # Edge case: if we are given no nodes, what do we do?
        if len(nodes) == 0:
            feature_size = self._node_attribute_arrays[node_type].shape[1]
            return np.empty((0, feature_size))

        # Get index for nodes of this type
        nt_id_to_index = self._node_index_maps[node_type]
        node_indices = [nt_id_to_index.get(n) for n in nodes]

        if None in node_indices:
            raise ValueError(
                "Incorrect node specified or nodes of multiple types found."
            )

        features = self._node_attribute_arrays[node_type][node_indices]
        return features

    def node_feature_sizes(self, node_types=None):
        """
        Get the feature sizes for the specified node types.

        Args:
            node_types: (list) A list of node types. If None all current node types
                will be used.

        Returns:
            A dictionary of node type and integer feature size.
        """
        # TODO: unit test!
        if not node_types:
            node_types = self.node_types

        self.check_graph_for_ml(features=True)

        fsize = {nt: self._node_attribute_arrays[nt].shape[1] for nt in node_types}
        return fsize

    def nodes_of_type(self, node_type=None):
        """
        Get the nodes of the graph with the specified node types.

        Args:
            node_type:

        Returns:
            A list of node IDs with type node_type
        """
        # TODO: unit test!
        if node_type is None:
            return list(self)
        else:
            return [
                n
                for n, ndata in self.nodes(data=True)
                if ndata.get(self._node_type_attr) == node_type
            ]

    @property
    def node_types(self):
        """
        Get a list of all node types in the graph.

        Returns:
            set of types
        """
        # TODO: unit test!
        # TODO: create a schmea when we geenrate _node_attribute_arrays and use it?
        if len(self._node_attribute_arrays) > 0:
            return set(self._node_attribute_arrays.keys())
        else:
            return {
                ndata.get(self._node_type_attr) for n, ndata in self.nodes(data=True)
            }

    def info(self, show_attributes=True, sample=None):
        """
        Return an information string summarizing information on the current graph.
        This includes node and edge type information and their attributes.

        Note: This requires processing all nodes and edges and could take a long
        time for a large graph.

        Args:
            sample (int): To speed up the graph analysis, use only a random sample of
                          this many nodes and edges.

        Returns:
            An information string.
        """
        directed_str = "Directed" if self.is_directed() else "Undirected"
        s = "{}: {} multigraph\n".format(type(self).__name__, directed_str)
        s += " Nodes: {}, Edges: {}\n".format(
            self.number_of_nodes(), self.number_of_edges()
        )

        # Sample the nodes for our analysis
        if sample:
            all_nodes = list(self.nodes)
            snodes = random.sample(all_nodes, sample)
            sedges = self.edges(snodes, keys=True)
        else:
            snodes = None
            sedges = None

        gs = self.create_graph_schema(create_type_maps=True, nodes=snodes, edges=sedges)

        # Go over all node types
        s += "\n Node types:\n"
        for nt in gs.node_types:
            # Filter nodes by type
            nt_nodes = [
                ndata for n, ndata in self.nodes(data=True) if gs.get_node_type(n) == nt
            ]
            s += "  {}: [{}]\n".format(nt, len(nt_nodes))

            # Get the attributes for this node type
            attrs = set(it.chain(*[ndata.keys() for ndata in nt_nodes]))
            attrs.discard(self._node_type_attr)
            if show_attributes and len(attrs) > 0:
                s += "        Attributes: {}\n".format(attrs)

            s += "    Edge types: "
            s += ", ".join(["{}-{}->{}".format(*e) for e in gs.schema[nt]]) + "\n"

        s += "\n Edge types:\n"
        for et in gs.edge_types:
            # Filter edges by type
            et_edges = [
                e[3]
                for e in self.edges(keys=True, data=True)
                if gs.get_edge_type(e[:3]) == et
            ]
            if len(et_edges) > 0:
                s += "    {et[0]}-{et[1]}->{et[2]}: [{len}]\n".format(
                    et=et, len=len(et_edges)
                )

            # Get the attributes for this edge type
            attrs = set(it.chain(*[edata.keys() for edata in et_edges]))
            attrs.discard(self._edge_type_attr)
            if show_attributes and len(attrs) > 0:
                s += "        Attributes: {}\n".format(attrs)

        return s

    def create_graph_schema(self, create_type_maps=True, nodes=None, edges=None):
        """
        Create graph schema in dict of dict format from current graph.

        Note the assumption we make that there is only one
        edge of a particular edge type per node pair.

        This means that specifying an edge by node0, node1 and edge type
        is unique.

        Returns:
            GraphSchema object.
        """
        if nodes is None:
            nodes = self.nodes()
        elif create_type_maps is True:
            raise ValueError(
                "Creating type mapes for subsampled nodes is not supported"
            )
        if edges is None:
            edges = self.edges(keys=True)
        elif create_type_maps is True:
            raise ValueError(
                "Creating type mapes for subsampled edges is not supported"
            )

        # Create node type index list
        node_types = sorted(
            {self.node[n].get(self._node_type_attr) for n in nodes}, key=str
        )

        if None in node_types:
            raise ValueError(
                "All nodes should have a type set in the '{}' attribute.".format(
                    self._node_type_attr
                )
            )

        graph_schema = {nt: set() for nt in node_types}

        # Create edge type index list
        edge_types = set()
        for n1, n2, k in edges:
            edata = self.adj[n1][n2][k]

            # Edge type tuple
            node_type_1 = self.node[n1][self._node_type_attr]
            node_type_2 = self.node[n2][self._node_type_attr]
            edge_type = edata[self._edge_type_attr]

            # Add edge type to node_type_1 data
            edge_type_tri = EdgeType(node_type_1, edge_type, node_type_2)
            edge_types.add(edge_type_tri)
            graph_schema[node_type_1].add(edge_type_tri)

            # Also add type to node_2 data if not digraph
            if not self.is_directed():
                edge_type_tri = EdgeType(node_type_2, edge_type, node_type_1)
                edge_types.add(edge_type_tri)
                graph_schema[node_type_2].add(edge_type_tri)

        # Create ordered list of edge_types
        edge_types = sorted(edge_types)

        # Create keys for node and edge types
        schema = {
            node_label: [
                edge_types[einx]
                for einx in sorted([edge_types.index(et) for et in list(node_data)])
            ]
            for node_label, node_data in graph_schema.items()
        }

        # Create schema object
        gs = GraphSchema()
        gs._is_directed = self.is_directed()
        gs.edge_types = edge_types
        gs.node_types = node_types
        gs.schema = schema

        # Create quick type lookups for nodes and edges.
        # Note: we encode the type index, in the assumption it will take
        # less storage.
        if create_type_maps:
            node_type_map = {
                n: node_types.index(ndata[self._node_type_attr])
                for n, ndata in self.nodes(data=True)
            }
            edge_type_map = {
                (edge[0], edge[1], edge[2]): edge_types.index(
                    EdgeType(
                        node_types[node_type_map[edge[0]]],
                        edge[3][self._edge_type_attr],
                        node_types[node_type_map[edge[1]]],
                    )
                )
                for edge in self.edges(keys=True, data=True)
            }

            gs.node_type_map = node_type_map
            gs.edge_type_map = edge_type_map

        return gs


class StellarGraph(StellarGraphBase, MultiGraph):
    def __init__(self, incoming_graph_data=None, **attr):
        super().__init__(incoming_graph_data, **attr)


class StellarDiGraph(StellarGraphBase, MultiDiGraph):
    def __init__(self, incoming_graph_data=None, **attr):
        super().__init__(incoming_graph_data, **attr)
